# -*- coding: utf-8 -*-
# Name: particle.py
# Authors: Stephan Meighen-Berger
# Containts the particle class definitions

import logging
from .config import config


_log = logging.getLogger(__name__)


class Particle(object):
    """Constructs the particle object.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Raises
    ------
    """
    def __init__(self, pdg_id: int):
        """Constructs the particle.

        Parameters
        ----------
        pdg_id : int
            The pdg number of the particle

        Returns
        -------
        None

        Raises
        ------
        """
        if not config["general"]["enable logging"]:
            _log.disabled = True
        _log.info('Constructing a particle')
        self._pdg_id = pdg_id
        # Naming conventions PDG Monte Carlo scheme
        self._name = config["pdg id"][pdg_id]
        _log.debug("The final name is " + self._name)
        self._energies = config["advanced"]["energy grid"]
        # TODO: Add masses for all particles
        # Masses of the muons
        if self._name[:2] == "mu":
            self._mass = config[self._name]["mass"]
            self._std_track = config[self._name]["standard track length"]



# -*- coding: utf-8 -*-
# Name: em_cascades.py
# Authors: Stephan Meighen-Berger
# Constructs an electromagnetic cascade, defined by the emitted photons

import logging
import numpy as np
import pickle
import pkgutil
from scipy.special import gamma as gamma_func
from .config import config
try:
    import jax.numpy as jnp
    from jax.scipy.stats import gamma as jax_gamma
except ImportError:
    if config["general"]["jax"]:
        raise ImportError("Jax not found!")

_log = logging.getLogger(__name__)


class EM_Cascade(object):
    """Constructs the em cascade object.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Raises
    ------
    """
    def __init__(self):
        """Constructs the em cascade.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Raises
        ------
        None
        """
        if not config["general"]["enable logging"]:
            _log.disabled = True
        _log.debug('Constructing an em cascade object')
        self._medium = config["mediums"][config["scenario"]["medium"]]
        self._n = self._medium["refractive index"]
        self._radlength = self._medium["radiation length"]
        self._Lrad = self._radlength / self._medium["density"]
        if config["scenario"]["parametrization"] == "aachen":
            _log.info("Loading the aachen parametrization")
            param_file = pkgutil.get_data(
                    __name__,
                    "data/%s.pkl" % config["scenario"]["parametrization"]
            )
            self._params = pickle.loads(param_file)["em cascade"]
        else:
            raise ValueError("EM parametrization " +
                             config["scenario"]["parametrization"] +
                             " not implemented!")
        if config["general"]["jax"]:
            _log.info("Running with JAX functions")
            self.cherenkov_angle_distro = self._symmetric_angle_distro_jax
            self.track_lengths = self._track_lengths_fetcher_jax
            self.long_profile = self._log_profile_func_fetcher_jax
        else:
            _log.info("Running with basic functions")
            self.cherenkov_angle_distro = self._symmetric_angle_distro
            self.track_lengths = self._track_lengths_fetcher
            self.long_profile = self._log_profile_func_fetcher

    ###########################################################################
    # Numpy
    def _track_lengths_fetcher(
            self, E, name: int):
        """ Parametrization for the energy dependence of the tracks. This is
        the fetcher function for a single energy

        Parameters
        ----------
        E : float/np.array
            The energy of interest
        name : int
            The particle of interest

        Returns
        -------
        track_length : np.array
            The track lengths for different energies
        track_length_dev : np.array
            The track lengths deviations for different energies
        """
        params = self._params["track parameters"][name]
        alpha = params["alpha"]
        beta = params["beta"]
        alpha_dev = params["alpha dev"]
        beta_dev = params["beta dev"]
        track_length = alpha * E**beta
        track_length_dev = alpha_dev * E**beta_dev
        return track_length, track_length_dev

    def _log_profile_func_fetcher(
            self, E, z: np.array, name: int,
            ) -> np.array:
        """ Parametrization of the longitudinal profile for a single energy.
        This still needs work

        Parameters
        ----------
        E : float/np.array
            The energy in GeV
        z : np.array
            The cascade depth in cm
        name : int
            The particle of interest

        Returns
        -------
        res : np.array
            Is equal to l^(-1) * dl/dt. The result will be 2 dimensional, with
            cm defined along the first axis and energies along the second
        """
        t = z / self._Lrad
        b = self._b_energy_fetch(name)
        a = self._a_energy_fetch(E, name)
        a = np.array(a).flatten()
        res = np.array([
            b * (
                (t * b)**(a_val - 1.) * np.exp(-(t*b)) / gamma_func(a_val)
            ) for a_val in a
        ])
        return res

    def _a_energy_fetch(self, E: float, name: int) -> np.array:
        """ Parametrizes the energy dependence of the a parameter for the
        longitudinal profiles. This is for a single energy.

        Parameters
        ----------
        E : float
            The energy in GeV
        name : int
            The particle of interest

        Returns
        -------
        a : np.array
            The values for the energies of interest
        """
        params = self._params["longitudinal parameters"][
            name]
        alpha = params["alpha"]
        beta = params["beta"]
        a = alpha + beta * np.log10(E)
        return a

    def _b_energy_fetch(self, name: int) -> np.array:
        """ Parametrizes the energy dependence of the b parameter for the
        longitudinal profiles. Currently assumed to be constant.
        This is for a single energy.

        Parameters
        ----------
        name : int
            The particle of interest

        Returns
        -------
        b : np.array
            The values for the energies of interest
        """
        params = self._params["longitudinal parameters"][
            name]
        b = params["b"]
        return b

    def _symmetric_angle_distro(
            self,
            phi: np.array, n: float,
            name: int) -> np.array:
        # TODO: Add asymmetry function
        # TODO: Add changes with shower depth
        """ Calculates the symmetric angular distribution of the Cherenkov
        emission. The error should lie below 10%

        Parameters
        ----------
        phi : np.array
            The angles of interest in degrees
        n : float
            The refractive index
        name : int
            The particle of interest

        Returns
        -------
        distro : np.array
            The distribution of emitted photons given the angle
        """
        params = self._params["angular distribution"][name]
        a = params["a"]
        b = params["b"]
        c = params["c"]
        d = params["d"]
        distro = (a * np.exp(b * np.abs(
            1. / n - np.cos(np.deg2rad(phi)))**c
        ) + d)
        return distro

    ###########################################################################
    # JAX
    def _track_lengths_fetcher_jax(
            self, E: float, name: int):
        """ Parametrization for the energy dependence of the tracks. This is
        the fetcher function for a single energy. JAX implementation

        Parameters
        ----------
        E : float
            The energy of interest
        name : int
            The particle of interest

        Returns
        -------
        track_length : float
            The track lengths for different energies
        track_length_dev : float
            The track lengths deviations for different energies
        """
        params = self._params["track parameters"][name]
        alpha = params["alpha"]
        beta = params["beta"]
        alpha_dev = params["alpha dev"]
        beta_dev = params["beta dev"]
        track_length = alpha * E**beta
        track_length_dev = alpha_dev * E**beta_dev
        return track_length, track_length_dev

    def _log_profile_func_fetcher_jax(
            self, E: float, z: float, name: int,
            ) -> float:
        """ Parametrization of the longitudinal profile for a single energy.
        This still needs work. JAX implementation

        Parameters
        ----------
        E : float
            The energy in GeV
        z : float
            The cascade depth in cm
        name : int
            The particle of interest

        Returns
        -------
        res : float
            Is equal to l^(-1) * dl/dt
        """
        t = z / self._Lrad
        b = self._b_energy_fetch_jax(name)
        a = self._a_energy_fetch_jax(E, name)
        res = jax_gamma.pdf(t * b, a) * b
        return res

    def _a_energy_fetch_jax(self, E: float, name: int) -> float:
        """ Parametrizes the energy dependence of the a parameter for the
        longitudinal profiles. This is for a single energy.

        Parameters
        ----------
        E : float
            The energy in GeV
        name : int
            The particle of interest

        Returns
        -------
        a : float
            The values for the energies of interest
        """
        params = self._params["longitudinal parameters"][
            name]
        alpha = params["alpha"]
        beta = params["beta"]
        a = alpha + beta * jnp.log10(E)
        return a

    def _b_energy_fetch_jax(self, name: int) -> float:
        """ Parametrizes the energy dependence of the b parameter for the
        longitudinal profiles. Currently assumed to be constant.
        This is for a single energy.

        Parameters
        ----------
        particle : Particle
            The particle of interest

        Returns
        -------
        b : float
            The values for the energies of interest
        """
        params = self._params["longitudinal parameters"][
            name]
        b = params["b"]
        return b

    def _symmetric_angle_distro_jax(
            self,
            phi: float, n: float,
            name: int) -> float:
        # TODO: Add asymmetry function
        # TODO: Add changes with shower depth
        """ Calculates the symmetric angular distribution of the Cherenkov
        emission. The error should lie below 10%

        Parameters
        ----------
        phi : float
            The angles of interest in degrees
        n : float
            The refractive index
        name : int
            The particle of interest

        Returns
        -------
        distro : float
            The distribution of emitted photons given the angle
        """
        params = self._params["angular distribution"][name]
        a = params["a"]
        b = params["b"]
        c = params["c"]
        d = params["d"]
        distro = (a * jnp.exp(b * jnp.abs(
            1. / n - jnp.cos(jnp.deg2rad(phi)))**c
        ) + d)
        return distro



# -*- coding: utf-8 -*-
# Name: photons.py
# Authors: Stephan Meighen-Berger
# Calculates the number of photons depending on the track length

import logging
import numpy as np
from .config import config
from .tracks import Track
from .em_cascades import EM_Cascade
from .hadron_cascades import Hadron_Cascade
try:
    import jax.numpy as jnp
    from jax import jit
    from jax.random import normal as jnormal
except ImportError:
    if config["general"]["jax"]:
        raise ImportError("Jax not found!")


_log = logging.getLogger(__name__)


class Photon(object):
    """Constructs the Photon object.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Raises
    ------
    """
    def __init__(
            self,
            particle, track: Track,
            em_cascade: EM_Cascade, hadron_cascade: Hadron_Cascade):
        """Constructs the photon.

        Parameters
        ----------
        particles : dic
            Library of the particles of interst, each being a Particle object
        track : Track
            The particle for which the tracks should be generated
        em_cascade : Particle
            The particle for which the tracks should be generated
        hadron_cascades : Hadron_Cascade
            The particle for which the tracks should be generated


        Returns
        -------
        None

        Raises
        ------
        ValueError
            Other distribution type for emission angles is not implemented
        """
        if not config["general"]["enable logging"]:
            _log.disabled = True
        _log.debug('Constructing a photon object')
        self._medium = config["scenario"]["medium"]
        self._n = config["mediums"][self._medium]["refractive index"]
        self._alpha = config["advanced"]["fine structure"]
        self._charge = config["advanced"]["particle charge"]
        self._wavelengths = config["advanced"]["wavelengths"]
        self._angle_grid = config["advanced"]["angles"]
        self._zgrid = config["advanced"]["z grid"]
        self.__particles = particle
        self.__track = track
        self.__em_cascade = em_cascade
        self.__hadron_cascade = hadron_cascade
        self._rstate = config["runtime"]["random state"]
        self._deltaL = config["advanced"]["track length"]
        self._track_interactions = config["simulation"]["track interactions"]
        # Building the functions
        _log.info("Building the necessary functions")
        self._track_builder()
        self._em_cascade_builder()
        self._hadron_cascade_builder()
        _log.debug('Finished a photon object.')

    ###########################################################################
    # The Fetchers
    def _track_fetcher(
            self, energy,
            wavelengths=config["advanced"]["wavelengths"],
            angle_grid=config["advanced"]["angles"],
            n=config["mediums"][
                config["scenario"]["medium"]]["refractive index"],
            interaction='total',
            function=False):
        """ Fetcher function for a specific energy and wavelength. This is for
        tracks and currently only for muons. Note in JAX mode the functions
        only take scalars!

        Parameters
        ----------
        energy : float
            The energy(ies) of the particle in GeV
        wavelengths : np.array
            Optional: The desired wavelengths
        angle_grid : np.array
            Optional: The desired angles
        n : float
            The refractive index of the medium.
        interaction : str
            Optional: The interaction which should produce the light
        function : bool
            Optional: returns the functional form instead of the evaluation

        Returns
        -------
        differential_counts : np.array/function
            dN/dlambda The differential photon counts per track length (in cm).
            The shape of the array is len(wavelengths).
        angles : np.array/function
            The angular distribution in degrees
        """
        if function:
            _log.debug("Fetching track functions for " + interaction)
            return (
                self._track_functions_dic[interaction]["dcounts"],
                self._track_functions_dic[interaction]["angles"]
            )
        else:
            _log.debug("Fetching track values for " + interaction)
            if config["general"]["jax"]:
                return(
                    np.array([
                        self._track_functions_dic[interaction]["dcounts"](
                            energy, wavelength
                        )
                        for wavelength in wavelengths
                    ]),
                    np.array([
                        self._track_functions_dic[interaction]["angles"](
                            angle,
                            n,
                            energy
                        )
                        for angle in angle_grid
                    ]),
                )
            else:
                return(
                    self._track_functions_dic[interaction]["dcounts"](
                        energy, wavelengths
                    ),
                    self._track_functions_dic[interaction]["angles"](
                        angle_grid,
                        n,
                        energy
                    )
                )

    def _em_cascade_fetcher(
            self, energy,
            particle: int,
            wavelengths=config["advanced"]["wavelengths"],
            angle_grid=config["advanced"]["angles"],
            n=config["mediums"][
                config["scenario"]["medium"]]["refractive index"],
            z_grid=config["advanced"]["z grid"],
            function=False):
        """ Fetcher function for a specific particle and energy. This is for
        em cascades.

       Parameters
        ----------
        energy : float
            The energy(ies) of the particle in GeV
        particle : int
            The pdg id of the particle of interest
        wavelengths : np.array
            Optional: The desired wavelengths
        angle_grid : np.array
            Optional: The desired angles in degress
        n : float
            Optional: The refractive index of the medium.
        z_grid : np.array
            Optional: The grid in cm for the long. distributions
        function : bool
            Optional: returns the functional form instead of the evaluation

        Returns
        -------
        differential_counts : function/float/np.array
            dN/dlambda The differential photon counts per track length (in cm).
            The shape of the array is (len(wavelengths), len(deltaL)).
        differential_counts_sample : float/np.array
            A sample of the differential counts distribution. Same shape as
            the differential counts
        long_profile : function/float/np.array
            The distribution along the shower axis for cm
        angles : function/float/np.array
            The angular distribution in degrees
        """
        if function:
            _log.debug("Fetching em functions for pdg_id " + str(particle))
            return (
                self._em_cascade_function_dic[particle]["dcounts"],
                self._em_cascade_function_dic[particle]["dcounts sample"],
                self._em_cascade_function_dic[particle]["long distro"],
                self._em_cascade_function_dic[particle]["angle distro"]
            )
        # Fetching explicit values
        else:
            _log.debug("Fetching track values for " + str(particle))
            if config["general"]["jax"]:
                return(
                    np.array([
                        self._em_cascade_function_dic[particle]["dcounts"](
                            energy, wavelength
                        )
                        for wavelength in wavelengths
                    ]),
                    np.array([
                        self._em_cascade_function_dic[particle][
                            "dcounts sample"
                            ](energy, wavelength)
                        for wavelength in wavelengths
                    ]),
                    np.array([
                        self._em_cascade_function_dic[particle][
                            "long distro"
                            ](energy, z)
                        for z in z_grid
                    ]),
                    np.array([
                        self._em_cascade_function_dic[particle][
                            "angle distro"](
                                angle,
                                n
                            )
                        for angle in angle_grid
                    ]),
                )
            else:
                return(
                        self._em_cascade_function_dic[particle]["dcounts"](
                            energy, wavelengths
                        ),
                        self._em_cascade_function_dic[particle][
                            "dcounts sample"
                            ](energy, wavelengths),
                        self._em_cascade_function_dic[particle][
                            "long distro"
                            ](energy, z_grid),
                        self._em_cascade_function_dic[particle][
                            "angle distro"](
                                angle_grid,
                                n
                            )
                )

    def _hadron_cascade_fetcher(
            self, energy,
            particle: int,
            wavelengths=config["advanced"]["wavelengths"],
            angle_grid=config["advanced"]["angles"],
            n=config["mediums"][
                config["scenario"]["medium"]]["refractive index"],
            z_grid=config["advanced"]["z grid"],
            function=False):
        """ Fetcher function for a specific particle and energy. This is for
        hadron cascades.

       Parameters
        ----------
        energy : float
            The energy(ies) of the particle in GeV
        particle : int
            The pdg id of the particle of interest
        wavelengths : np.array
            Optional: The desired wavelengths
        angle_grid : np.array
            Optional: The desired angles in degress
        n : float
            Optional: The refractive index of the medium.
        z_grid : np.array
            Optional: The grid in cm for the long. distributions
        function : bool
            Optional: returns the functional form instead of the evaluation

        Returns
        -------
        differential_counts : function/float/np.array
            dN/dlambda The differential photon counts per track length (in cm).
            The shape of the array is (len(wavelengths), len(deltaL)).
        differential_counts_sample : float/np.array
            A sample of the differential counts distribution. Same shape as
            the differential counts
        em_fraction_mean : float/np.array
            The fraction of em particles
        em_fraction_sample : float/np.array
            A sample of the em_fraction
        long_profile : function/float/np.array
            The distribution along the shower axis for cm
        angles : function/float/np.array
            The angular distribution in degrees
        """
        if function:
            _log.debug("Fetching em functions for pdg_id " + str(particle))
            return (
                self._hadron_cascade_function_dic[particle]["dcounts"],
                self._hadron_cascade_function_dic[particle]["dcounts sample"],
                self._hadron_cascade_function_dic[particle][
                    "em fraction mean"
                ],
                self._hadron_cascade_function_dic[particle][
                    "em fraction sample"
                ],
                self._hadron_cascade_function_dic[particle]["long distro"],
                self._hadron_cascade_function_dic[particle]["angle distro"]
            )
        # Fetching explicit values
        else:
            _log.debug("Fetching track values for " + str(particle))
            if config["general"]["jax"]:
                return(
                    np.array([
                        self._hadron_cascade_function_dic[particle]["dcounts"](
                            energy, wavelength
                        )
                        for wavelength in wavelengths
                    ]),
                    np.array([
                        self._hadron_cascade_function_dic[particle][
                            "dcounts sample"
                            ](energy, wavelength)
                        for wavelength in wavelengths
                    ]),
                    (
                        self._hadron_cascade_function_dic[particle][
                            "em fraction mean"
                            ](energy)
                    ),
                    (
                        self._hadron_cascade_function_dic[particle][
                            "em fraction sample"
                            ](energy)
                    ),
                    np.array([
                        self._hadron_cascade_function_dic[particle][
                            "long distro"
                            ](energy, z)
                        for z in z_grid
                    ]),
                    np.array([
                        self._hadron_cascade_function_dic[particle][
                            "angle distro"](
                                energy,
                                angle,
                                n
                            )
                        for angle in angle_grid
                    ]),
                )
            else:
                return(
                        self._hadron_cascade_function_dic[particle]["dcounts"](
                            energy, wavelengths
                        ),
                        self._hadron_cascade_function_dic[particle][
                            "dcounts sample"
                            ](energy, wavelengths),
                        self._hadron_cascade_function_dic[particle][
                            "em fraction mean"
                            ](energy),
                        self._hadron_cascade_function_dic[particle][
                            "em fraction sample"
                            ](energy),
                        self._hadron_cascade_function_dic[particle][
                            "long distro"
                            ](energy, z_grid),
                        self._hadron_cascade_function_dic[particle][
                            "angle distro"](
                                energy,
                                angle_grid,
                                n
                            )
                )
    ###########################################################################
    # The Builders

    def _track_builder(self):
        """ Builder function for the track functions.

        Parameters
        ----------
        interaction : str
            Optional: The interaction(s) which should produce the light

        Returns
        -------
        None
        """
        # Looping over the interaction types
        _log.info("Building track functions")
        self._track_functions_dic = {}
        for interaction in self._track_interactions:
            # Photon count function and angle function
            self._track_functions_dic[interaction] = {}

            # Building the counts functionn
            def track_mean(energy, interaction=interaction):
                """ Fetcher function for a specific particle
                and energy. This is for tracks

                Parameters
                ----------
                energy : float
                    The energy of the particle
                interaction : str
                    The interaction type

                Returns
                -------
                counts : float
                    The photon counts
                """
                tmp_track_frac = (
                    self.__track.additional_track_ratio(
                        energy, interaction=interaction
                    )
                )
                new_track = self._deltaL * (1. + tmp_track_frac)
                return new_track
            if config["general"]["jax"]:
                _log.debug(
                    "Constructing Jax function for " + interaction
                )

                def counts_mean(energy, wavelengths, interaction=interaction):
                    """ Calculates the differential photon counts.
                    Jax implemenation
                    Parameters
                    ----------
                    energy : float
                        The energy of the particle in GeV
                    wavelengths: float/np.array
                        The wavelength(s) of interest
                    interaction : str
                        The interaction type

                    Returns
                    -------
                    differential_counts : np.array
                        The differential photon counts
                        per track length (in cm).
                    """
                    new_track = track_mean(energy, interaction=interaction)
                    return self._cherenkov_counts_jax(wavelengths, new_track)
                # jitting
                counts = jit(counts_mean, static_argnames=['interactions'])
                angles = jit(self.__track.cherenkov_angle_distro)
            else:
                _log.debug(
                    "Constructing Jax function for " + interaction
                )

                def counts_mean(energy, wavelengths, interaction=interaction):
                    """Calculates the differential photon counts.

                    Parameters
                    ----------
                    energy : float
                        The energy of the particle in GeV
                    wavelengths: float/np.array
                        The wavelength(s) of interest
                    interaction : str
                        The interaction type

                    Returns
                    -------
                    differential_counts : np.array
                        The differential photon counts
                        per track length (in cm).
                    """
                    new_track = track_mean(energy, interaction=interaction)
                    return self._cherenkov_counts(wavelengths, new_track)
                # No jitting here
                counts = counts_mean
                angles = self.__track.cherenkov_angle_distro
            self._track_functions_dic[interaction]["dcounts"] = counts
            self._track_functions_dic[interaction]["angles"] = angles

    def _em_cascade_builder(self):
        """ Builder function for a the cascade functions. This is for
        em cascades.

        Parameters
        ----------

        Returns
        -------
        None
        """
        _log.debug("Building the em cascade functions")
        self._em_cascade_function_dic = {}
        for particle_id in config["simulation"]["em particles"]:
            name = particle_id
            self._em_cascade_function_dic[particle_id] = {}

            def track_mean(energy, name=name):
                """ Fetcher function for a specific particle and energy.
                This is for em cascades and their photon counts

                Parameters
                ----------
                energy : float
                    The energy of the particle
                name : int
                    The name of the particle of interest

                Returns
                -------
                tmp_track : float
                    The track length
                """
                tmp_track, _ = (
                    self.__em_cascade.track_lengths(
                        energy, name
                    )
                )
                return tmp_track

            def track_sampler(energy, name=name):
                """ Fetcher function for a specific particle and energy.
                This samples the distribution

                Parameters
                ----------
                energy : float
                    The energy of the particle
                name : int
                    The name of the particle of interest

                Returns
                -------
                tmp_track_sample : float
                    The sampled photon counts
                """
                tmp_track, tmp_track_sd = (
                    self.__em_cascade.track_lengths(
                        energy, name
                    )
                )
                if config["general"]["jax"]:
                    tmp_track_sample = (
                        tmp_track + tmp_track_sd * jnormal(self._rstate)
                    )
                else:
                    tmp_track_sample = self._rstate.normal(
                        tmp_track, tmp_track_sd
                    )
                return tmp_track_sample

            def long_profile(energy, z_grid, name=name):
                """ The longitudinal profile of the em cascade

                Parameters
                ----------
                energy : float
                    The energy of the particle
                z_grid : float/np.array
                    The grid to evaluate the distribution in in cm
                name : int
                    The name of the particle of interest

                Returns
                -------
                long_distro : float/np.array
                    The resulting longitudinal distribution
                """
                return self.__em_cascade.long_profile(energy, z_grid, name)

            def angle_distro(
                    angles,
                    n=config["mediums"][self._medium]["refractive index"],
                    name=name
                    ):
                """ The angle distribution of the cherenkov photons for
                the em cascade

                Parameters
                ----------
                angles : float/np.array
                    The angles of interest
                n : float
                    Optional: The refractive index of the material
                name : int
                    Optional: The name of the particle of interest

                Returns
                -------
                angle_distro : float/np.array
                    The resulting longitudinal distribution
                """
                return self.__em_cascade.cherenkov_angle_distro(
                    angles, n, name
                )
            # Storing the functions
            if config["general"]["jax"]:
                _log.debug(
                    "Constructing Jax function for pdg_id " + str(name)
                )

                def counts_mean(energy, wavelengths, name=name):
                    """Calculates the differential photon counts.
                    Jax implemenation

                    Parameters
                    ----------
                    energy : float
                        The energy of the particle in GeV
                    wavelengths: float
                        The wavelength(s) of interest
                    name : int
                        Name of the particle

                    Returns
                    -------
                    differential_counts : float
                        The differential photon counts
                        per track length (in cm).
                    """
                    new_track = track_mean(energy, name=name)
                    return self._cherenkov_counts_jax(wavelengths, new_track)

                def counts_sampler(energy, wavelengths, name=name):
                    """Calculates a sample of the differential photon counts.
                    Jax implemenation

                    Parameters
                    ----------
                    energy : float
                        The energy of the particle in GeV
                    wavelengths: float
                        The wavelength(s) of interest
                    name : int
                        Name of the particle

                    Returns
                    -------
                    differential_counts : float
                        The differential photon counts
                        per track length (in cm).
                    """
                    new_track = track_sampler(energy, name=name)
                    return self._cherenkov_counts_jax(wavelengths, new_track)
                # Jit the jax functions
                counts = jit(counts_mean, static_argnames=['name'])
                counts_sample = jit(counts_sampler, static_argnames=['name'])
                long = jit(long_profile, static_argnames=['name'])
                angles = jit(angle_distro, static_argnames=['name'])
            else:
                _log.debug(
                    "Constructing numpy function for pdg_id " + str(name)
                )

                def counts_mean(energy, wavelengths, name=name):
                    """Calculates the differential photon counts.
                    Jax implemenation

                    Parameters
                    ----------
                    energy : float
                        The energy of the particle in GeV
                    wavelengths: np.array
                        The wavelength(s) of interest
                    name : int
                        Name of the particle

                    Returns
                    -------
                    differential_counts : np.array
                        The differential photon counts
                        per track length (in cm).
                    """
                    new_track = track_mean(energy, name=name)
                    return self._cherenkov_counts(wavelengths, new_track)

                def counts_sampler(energy, wavelengths, name=name):
                    """Calculates a sample of the differential photon counts.
                    Jax implemenation

                    Parameters
                    ----------
                    energy : float
                        The energy of the particle in GeV
                    wavelengths: np.array
                        The wavelength(s) of interest
                    name : int
                        Name of the particle

                    Returns
                    -------
                    differential_counts : np.array
                        The differential photon counts
                        per track length (in cm).
                    """
                    new_track = track_sampler(energy, name=name)
                    return self._cherenkov_counts(wavelengths, new_track)
                # Don't jist the numpy functions
                counts = counts_mean
                counts_sample = counts_sampler
                long = long_profile
                angles = angle_distro
            # Storing
            self._em_cascade_function_dic[particle_id] = {
                "dcounts": counts,
                "dcounts sample": counts_sample,
                "long distro": long,
                "angle distro": angles,
            }

    def _hadron_cascade_builder(self):
        """ Builder function for a hadronic cascades.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        _log.debug("Building the hadron cascade functions")
        self._hadron_cascade_function_dic = {}
        for particle_id in config["simulation"]["hadron particles"]:
            name = particle_id
            self._hadron_cascade_function_dic[particle_id] = {}

            def track_mean(energy, name=name):
                """ Fetcher function for a specific particle and energy.
                This is for hadron cascades and their photon counts

                Parameters
                ----------
                energy : float
                    The energy of the particle
                name : int
                    The name of the particle of interest

                Returns
                -------
                tmp_track : float
                    The track length
                """
                tmp_track, _ = (
                    self.__hadron_cascade.track_lengths(
                        energy, name
                    )
                )
                return tmp_track

            def track_sampler(energy, name=name):
                """ Fetcher function for a specific particle and energy.
                This samples the distribution

                Parameters
                ----------
                energy : float
                    The energy of the particle
                name : int
                    The name of the particle of interest

                Returns
                -------
                tmp_track_sample : float
                    The sampled photon counts
                """
                tmp_track, tmp_track_sd = (
                    self.__hadron_cascade.track_lengths(
                        energy, name
                    )
                )
                if config["general"]["jax"]:
                    tmp_track_sample = (
                        tmp_track + tmp_track_sd * jnormal(self._rstate)
                    )
                else:
                    tmp_track_sample = self._rstate.normal(
                        tmp_track, tmp_track_sd
                    )
                return tmp_track_sample

            def em_fraction_mean(energy, name=name):
                """ The em fraction mean of the hadron cascade

                Parameters
                ----------
                energy : float
                    The energy of the particle
                name : int
                    The name of the particle of interest

                Returns
                -------
                em_frac_mean : float/np.array
                    The resulting em fraction
                """
                em_frac_mean, _ = (
                    self.__hadron_cascade.em_fraction(energy, name)
                )
                return em_frac_mean

            def em_fraction_sampler(energy, name=name):
                """ The em fraction sample of the hadron cascade

                Parameters
                ----------
                energy : float
                    The energy of the particle
                name : int
                    The name of the particle of interest

                Returns
                -------
                em_frac_sample : float/np.array
                    The resulting em fraction
                """
                em_frac_mean, em_frac_std = (
                    self.__hadron_cascade.em_fraction(energy, name)
                )
                if config["general"]["jax"]:
                    em_frac_sample = (
                        em_frac_mean + em_frac_std * jnormal(self._rstate)
                    )
                else:
                    em_frac_sample = self._rstate.normal(
                        em_frac_mean, em_frac_std
                    )
                return em_frac_sample

            def long_profile(energy, z_grid, name=name):
                """ The longitudinal profile of the hadron cascade

                Parameters
                ----------
                energy : float
                    The energy of the particle
                z_grid : float/np.array
                    The grid to evaluate the distribution in in cm
                name : int
                    The name of the particle of interest

                Returns
                -------
                long_distro : float/np.array
                    The resulting longitudinal distribution
                """
                return self.__hadron_cascade.long_profile(energy, z_grid, name)

            def angle_distro(
                    energy,
                    angles,
                    n=config["mediums"][self._medium]["refractive index"],
                    name=name
                    ):
                """ The angle distribution of the cherenkov photons for
                the hadron cascade

                Parameters
                ----------
                angles : float/np.array
                    The angles of interest
                n : float
                    Optional: The refractive index of the material
                name : int
                    Optional: The name of the particle of interest

                Returns
                -------
                angle_distro : float/np.array
                    The resulting longitudinal distribution
                """
                return self.__hadron_cascade.cherenkov_angle_distro(
                    energy, angles, n, name
                )
            # Storing the functions
            if config["general"]["jax"]:
                _log.debug(
                    "Constructing Jax function for pdg_id " + str(name)
                )

                def counts_mean(energy, wavelengths, name=name):
                    """Calculates the differential photon counts.
                    Jax implemenation

                    Parameters
                    ----------
                    energy : float
                        The energy of the particle in GeV
                    wavelengths: float
                        The wavelength(s) of interest
                    name : int
                        Name of the particle

                    Returns
                    -------
                    differential_counts : float
                        The differential photon counts
                        per track length (in cm).
                    """
                    new_track = track_mean(energy, name=name)
                    return self._cherenkov_counts_jax(wavelengths, new_track)

                def counts_sampler(energy, wavelengths, name=name):
                    """Calculates a sample of the differential photon counts.
                    Jax implemenation

                    Parameters
                    ----------
                    energy : float
                        The energy of the particle in GeV
                    wavelengths: float
                        The wavelength(s) of interest
                    name : int
                        Name of the particle

                    Returns
                    -------
                    differential_counts : float
                        The differential photon counts
                        per track length (in cm).
                    """
                    new_track = track_sampler(energy, name=name)
                    return self._cherenkov_counts_jax(wavelengths, new_track)
                # Jit the jax functions
                counts = jit(counts_mean, static_argnames=['name'])
                counts_sample = jit(counts_sampler, static_argnames=['name'])
                em_frac_mean = jit(em_fraction_mean, static_argnames=['name'])
                em_frac_sample = jit(
                    em_fraction_sampler,
                    static_argnames=['name']
                )
                long = jit(long_profile, static_argnames=['name'])
                angles = jit(angle_distro, static_argnames=['name'])
            else:
                _log.debug(
                    "Constructing numpy function for pdg_id " + str(name)
                )

                def counts_mean(energy, wavelengths, name=name):
                    """Calculates the differential photon counts.
                    Jax implemenation

                    Parameters
                    ----------
                    energy : float
                        The energy of the particle in GeV
                    wavelengths: np.array
                        The wavelength(s) of interest
                    name : int
                        Name of the particle

                    Returns
                    -------
                    differential_counts : np.array
                        The differential photon counts
                        per track length (in cm).
                    """
                    new_track = track_mean(energy, name=name)
                    return self._cherenkov_counts(wavelengths, new_track)

                def counts_sampler(energy, wavelengths, name=name):
                    """Calculates a sample of the differential photon counts.
                    Jax implemenation

                    Parameters
                    ----------
                    energy : float
                        The energy of the particle in GeV
                    wavelengths: np.array
                        The wavelength(s) of interest
                    name : int
                        Name of the particle

                    Returns
                    -------
                    differential_counts : np.array
                        The differential photon counts
                        per track length (in cm).
                    """
                    new_track = track_sampler(energy, name=name)
                    return self._cherenkov_counts(wavelengths, new_track)
                # Don't jist the numpy functions
                counts = counts_mean
                counts_sample = counts_sampler
                em_frac_mean = em_fraction_mean
                em_frac_sample = em_fraction_sampler
                long = long_profile
                angles = angle_distro
            # Storing
            self._hadron_cascade_function_dic[particle_id] = {
                "dcounts": counts,
                "dcounts sample": counts_sample,
                "em fraction mean": em_frac_mean,
                "em fraction sample": em_frac_sample,
                "long distro": long,
                "angle distro": angles,
            }

    def _cherenkov_counts(
            self,
            wavelengths: np.array, track_length: float) -> np.array:
        """ Calculates the differential number of photons for the given
        wavelengths and track-lengths assuming a constant velocity with beta=1.

        Parameters
        ----------
        wavelengths : np.array
            The wavelengths of interest
        track_lengths : float
            The track lengths of interest in cm

        Returns
        -------
        counts : np.array
            A array filled witht the produced photons.
        """
        prefac = (
            2. * np.pi * self._alpha * self._charge**2. /
            (1. - 1. / self._n**2.)
        )
        # 1e-7 due to the conversion from nm to cm
        diff_counts = np.array([
            prefac / (lambd * 1e-9)**2. * track_length * 1e-2
            for lambd in wavelengths
        ])
        return diff_counts * 1e-9 / np.pi

    def _cherenkov_counts_jax(
            self,
            wavelengths: float, track_lengths: float) -> float:
        """ Calculates the differential number of photons for the given
        wavelengths and track-lengths assuming a constant velocity with beta=1.

        Parameters
        ----------
        wavelengths : float
            The wavelengths of interest
        track_lengths : float
            The track lengths of interest in cm

        Returns
        -------
        counts : float
            The counts (differential)
        """
        prefac = (
            2. * jnp.pi * self._alpha * self._charge**2. /
            (1. - 1. / self._n**2.)
        )
        # 1e-7 due to the conversion from nm to cm
        diff_counts = (
            prefac / (wavelengths * 1e-9)**2. * track_lengths * 1e-2
        )
        return diff_counts * 1e-9 / jnp.pi



# -*- coding: utf-8 -*-
# __main__.py
# Authors: Stephan Meighen-Berger

def main():
    """ Tells the user how to use fennel
    """
    print("Welcome to fennel!")
    print("Please note fennel is a package and cannot be directly executed")
    print("The current fennel version is 1.2.2")
    print("Usage after installation:")
    print("from fennel import Fennel")
    print("fennel_obj = Fennel()")
    print("Visit the following link for examples:")
    print("https://github.com/MeighenBergerS/fennel/tree/master/notebooks")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Name: fennel.py
# Authors: Stephan Meighen-Berger
# Main interface to the fennel model. Calculates the light yields using the
# Aachen parametrization from
# https://www.institut3b.physik.rwth-aachen.de/global/show_document.asp?id=aaaaaaaaaapwhjz
# which is Leif Raedel's Master thesis

# Imports
# Native modules
import logging
import sys
import numpy as np
import yaml
# -----------------------------------------
# Package modules
from .config import config
from .particle import Particle
from .tracks import Track
from .em_cascades import EM_Cascade
from .hadron_cascades import Hadron_Cascade
from .photons import Photon
from .definition_generator import Definitions_Generator
try:
    from jax.random import PRNGKey
except ImportError:
    if config["general"]["jax"]:
        raise ImportError("Jax not found!")

# unless we put this class in __init__, __name__ will be contagion.contagion
_log = logging.getLogger("fennel")


class Fennel(object):
    """
    class: Fennel
    Interace to the fennel package. This class
    stores all methods required to run the simulation
    of the particle light yields
    Parameters
    ----------
    config : dic
        Configuration dictionary for the simulation

    Returns
    -------
    None
    """
    def __init__(self, userconfig=None):
        """
        function: __init__
        Initializes the class fennel.
        Here all run parameters are set.
        Parameters
        ----------
        config : dic
            Configuration dictionary for the simulation

        Returns
        -------
        None
        """
        # Inputs
        if userconfig is not None:
            if isinstance(userconfig, dict):
                config.from_dict(userconfig)
            else:
                config.from_yaml(userconfig)

        # Create RandomState
        if config["general"]["random state seed"] is None:
            _log.warning("No random state seed given, constructing new state")
            if config["general"]["jax"]:
                rstate = PRNGKey(1337)
            else:
                rstate = np.random.RandomState()
        else:
            if config["general"]["jax"]:
                rstate = PRNGKey(config["general"]["random state seed"])
            else:
                rstate = np.random.RandomState(
                    config["general"]["random state seed"]
                )
        config["runtime"] = {"random state": rstate}

        # Logger
        # Logging formatter
        fmt = "%(levelname)s: %(message)s"
        fmt_with_name = "[%(name)s] " + fmt
        formatter_with_name = logging.Formatter(fmt=fmt_with_name)
        # creating file handler with debug messages
        if config["general"]["enable logging"]:
            fh = logging.FileHandler(
                config["general"]["log file handler"], mode="w"
            )
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(formatter_with_name)
            _log.addHandler(fh)
        else:
            _log.disabled = True
        # console logger with a higher log level
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(config["general"]["debug level"])
        # add class name to ch only when debugging
        if config["general"]["debug level"] == logging.DEBUG:
            ch.setFormatter(formatter_with_name)
        else:
            formatter = logging.Formatter(fmt=fmt)
            ch.setFormatter(formatter)
        _log.addHandler(ch)
        _log.setLevel(logging.DEBUG)
        _log.info('---------------------------------------------------')
        _log.info('---------------------------------------------------')
        _log.info('Welcome to Fennel!')
        _log.info('This package will help you model light yields')
        _log.info('---------------------------------------------------')
        _log.info('---------------------------------------------------')
        _log.info('Creating particles...')
        self._particles = {}
        for particle_id in config["pdg id"].keys():
            # Particle creation
            self._particles[particle_id] = Particle(particle_id)
        _log.info('Creation finished')
        _log.info('---------------------------------------------------')
        _log.info('---------------------------------------------------')
        _log.info('Creating a track...')
        # Track creation
        self._track = Track()
        _log.info('Creation finished')
        _log.info('---------------------------------------------------')
        _log.info('---------------------------------------------------')
        _log.info('Creating an em cascade...')
        # EM cascade creation
        self._em_cascade = EM_Cascade()
        _log.info('Creation finished')
        _log.info('---------------------------------------------------')
        _log.info('---------------------------------------------------')
        _log.info('Creating a hadron cascade...')
        # Hadron cascade creation
        self._hadron_cascade = Hadron_Cascade()
        _log.info('Creation finished')
        _log.info('---------------------------------------------------')
        _log.info('---------------------------------------------------')
        _log.info('Creating a photon...')
        # Hadron cascade creation
        self._photon = Photon(
            self._particles, self._track,
            self._em_cascade, self._hadron_cascade
        )
        # Creating the definitions storer
        self._dg = Definitions_Generator(
            self._track, self._em_cascade, self._hadron_cascade
        )
        _log.info('Creation finished')
        _log.info('---------------------------------------------------')
        _log.info('---------------------------------------------------')

    def close(self):
        """ Wraps up the program

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        _log.info('---------------------------------------------------')
        _log.info('---------------------------------------------------')
        # A new simulation
        if config["general"]["enable logging"]:
            _log.debug(
                "Dumping run settings into %s",
                config["general"]["config location"],
            )
            with open(config["general"]["config location"], "w") as f:
                yaml.dump(config, f)
            _log.debug("Finished dump")
        _log.info('---------------------------------------------------')
        _log.info('---------------------------------------------------')
        _log.info("Have a great day and until next time!")
        _log.info("                  @*****&@         @@.                    ")
        _log.info("           @@@((@ @*******@     @%((((@@((@               ")
        _log.info("         @(((((((@(@*******@@@((@(((((((((@*              ")
        _log.info("        @((((((@(((@*******@(((@%(((((&%(((@              ")
        _log.info("         #@((@((&@((&*******@((@(((((@#(((@               ")
        _log.info("        @@*****@((#@@********@@(((((@(((@/**@@            ")
        _log.info("        @********@((@@*************@@@#@******&%          ")
        _log.info("@@  @ @&(@(***,*******,*******,*******,*******,@      .@@*")
        _log.info(" @   @@((((@*************.*********...*******@(@      ./  ")
        _log.info(" @     %(((#@**********...*,.. ........**,**(@(#*     @   ")
        _log.info("  @      @@,.*.***,@@@.......*,...@@@..*,.....@@    &@    ")
        _log.info("   @(    @....,,...@@......... ,*.@%...*......,%  @@      ")
        _log.info("       @@@.....,***...............,****.......*@,         ")
        _log.info("        .........*............ ........****..*.@          ")
        _log.info("         #*........*...&       @...............@          ")
        _log.info("         @ *.. ... ..,,..@@@@. ... ... ... ...&           ")
        _log.info("          @.,*...........*,...................@           ")
        _log.info("            @.**.............*,..............@            ")
        _log.info("             .@.**...............**.........@             ")
        _log.info("           @@@%   &@@,........ .......**#@    @@          ")
        _log.info("        @,                 (@@@@@@@@(             @       ")
        _log.info("                                                %@*       ")
        _log.info('---------------------------------------------------')
        _log.info('---------------------------------------------------')
        # Closing log
        logging.shutdown()

    def auto_yields(
            self, energy,
            particle: int,
            interaction='total',
            wavelengths=config["advanced"]["wavelengths"],
            angle_grid=config["advanced"]["angles"],
            n=config["mediums"][
                config["scenario"]["medium"]]["refractive index"],
            z_grid=config["advanced"]["z grid"],
            function=False):
        """ Auto fetcher function for a given particle and energy. This will
        fetch/evaluate the functions corresponding to the given particle.
        Some of the output will be none depending on the constructed object

        Parameters
        ----------
        energy : float
            The energy(ies) of the particle in GeV
        particle : int
            The pdg id of the particle of interest
        wavelengths : np.array
            Optional: The desired wavelengths
        interaction : str
            Optional: The interaction which should produce the light.
            This is used during track construction.
        angle_grid : np.array
            Optional: The desired angles in degress
        n : float
            Optional: The refractive index of the medium.
        z_grid : np.array
            Optional: The grid in cm for the long. distributions.
            Used when modeling cascades.
        function : bool
            Optional: returns the functional form instead of the evaluation

        Returns
        -------
        differential_counts : function/float/np.array
            dN/dlambda The differential photon counts per track length (in cm).
            The shape of the array is (len(wavelengths), len(deltaL)).
        differential_counts_sample : float/np.array
            A sample of the differential counts distribution. Same shape as
            the differential counts
        em_fraction_mean : float/np.array
            The fraction of em particles
        em_fraction_sample : float/np.array
            A sample of the em_fraction
        long_profile : function/float/np.array
            The distribution along the shower axis for cm
        angles : function/float/np.array
            The angular distribution in degrees
        """
        if particle in config["simulation"]["track particles"]:
            _log.debug("Fetching/evaluating track functions for " +
                       str(particle))
            dcounts, angles = self.track_yields(
                energy,
                wavelengths=wavelengths,
                angle_grid=angle_grid,
                n=n,
                interaction=interaction,
                function=function
            )
            # Unfilled variables
            dcounts_s = None
            em_frac = None
            em_frac_s = None
            long = None
        elif particle in config["simulation"]["em particles"]:
            _log.debug("Fetching/evaluating em functions for " +
                       str(particle))
            dcounts, dcounts_s, long, angles = self.em_yields(
                energy,
                particle,
                wavelengths=wavelengths,
                angle_grid=angle_grid,
                n=n,
                z_grid=z_grid,
                function=function
            )
            # Unfilled variables
            em_frac = None
            em_frac_s = None
        elif particle in config["simulation"]["hadron particles"]:
            _log.debug("Fetching/evaluating hadron functions for " +
                       str(particle))
            dcounts, dcounts_s, em_frac, em_frac_s, long, angles = (
                self.hadron_yields(
                    energy,
                    particle,
                    wavelengths=wavelengths,
                    angle_grid=angle_grid,
                    n=n,
                    z_grid=z_grid,
                    function=function
                )
            )
        else:
            raise ValueError(
                "Track/cascade object corresponding to " +
                str(particle) + " is unknown. Please contact " +
                "the authors if there is a need for this species"
            )
        return dcounts, dcounts_s, em_frac, em_frac_s, long, angles

    def track_yields(
            self,
            energy: float,
            wavelengths=config["advanced"]["wavelengths"],
            angle_grid=config["advanced"]["angles"],
            n=config["mediums"][
                config["scenario"]["medium"]]["refractive index"],
            interaction='total',
            function=False):
        """ Fetcher function for a specific energy and wavelength. This is for
        tracks and currently only for muons. Note in JAX mode the functions
        only take scalars!

        Parameters
        ----------
        energy : float
            The energy(ies) of the particle in GeV
        wavelengths : np.array
            Optional: The desired wavelengths
        angle_grid : np.array
            Optional: The desired angles
        n : float
            The refractive index of the medium.
        interaction : str
            Optional: The interaction which should produce the light
        function : bool
            returns the functional form instead of the evaluation

        Returns
        -------
        differential_counts : np.array/function
            dN/dlambda The differential photon counts per track length (in cm).
            The shape of the array is len(wavelengths).
        angles : np.array/function
            The angular distribution in degrees
        """
        return self._photon._track_fetcher(
            energy, wavelengths, angle_grid, n, interaction, function
        )

    def em_yields(
            self, energy,
            particle: int,
            wavelengths=config["advanced"]["wavelengths"],
            angle_grid=config["advanced"]["angles"],
            n=config["mediums"][
                config["scenario"]["medium"]]["refractive index"],
            z_grid=config["advanced"]["z grid"],
            function=False):
        """ Fetcher function for a specific particle and energy. This is for
        em cascades.

        Parameters
        ----------
        energy : float
            The energy(ies) of the particle in GeV
        particle : int
            The pdg id of the particle of interest
        wavelengths : np.array
            Optional: The desired wavelengths
        angle_grid : np.array
            Optional: The desired angles in degress
        n : float
            Optional: The refractive index of the medium.
        z_grid : np.array
            Optional: The grid in cm for the long. distributions
        function : bool
            Optional: returns the functional form instead of the evaluation

        Returns
        -------
        differential_counts : function/float/np.array
            dN/dlambda The differential photon counts per track length (in cm).
            The shape of the array is (len(wavelengths), len(deltaL)).
        differential_counts_sample : float/np.array
            A sample of the differential counts distribution. Same shape as
            the differential counts
        long_profile : function/float/np.array
            The distribution along the shower axis for cm
        angles : function/float/np.array
            The angular distribution in degrees
        """
        return self._photon._em_cascade_fetcher(
            energy, particle, wavelengths, angle_grid, n, z_grid, function
        )

    def hadron_yields(
            self, energy,
            particle: int,
            wavelengths=config["advanced"]["wavelengths"],
            angle_grid=config["advanced"]["angles"],
            n=config["mediums"][
                config["scenario"]["medium"]]["refractive index"],
            z_grid=config["advanced"]["z grid"],
            function=False):
        """ Fetcher function for a specific particle and energy. This is for
        hadron cascades.

        Parameters
        ----------
        energy : float
            The energy(ies) of the particle in GeV
        particle : int
            The pdg id of the particle of interest
        wavelengths : np.array
            Optional: The desired wavelengths
        angle_grid : np.array
            Optional: The desired angles in degress
        n : float
            Optional: The refractive index of the medium.
        z_grid : np.array
            Optional: The grid in cm for the long. distributions
        function : bool
            Optional: returns the functional form instead of the evaluation

        Returns
        -------
        differential_counts : function/float/np.array
            dN/dlambda The differential photon counts per track length (in cm).
            The shape of the array is (len(wavelengths), len(deltaL)).
        differential_counts_sample : float/np.array
            A sample of the differential counts distribution. Same shape as
            the differential counts
        em_fraction_mean : float/np.array
            The fraction of em particles
        em_fraction_sample : float/np.array
            A sample of the em_fraction
        long_profile : function/float/np.array
            The distribution along the shower axis for cm
        angles : function/float/np.array
            The angular distribution in degrees
        """
        return self._photon._hadron_cascade_fetcher(
            energy, particle, wavelengths, angle_grid, n, z_grid, function
        )

    def definitions(self):
        """ Write the definitions file

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        self._dg._write()

    def pars2csv(self):
        """ Write the parameters to a csv file

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        self._dg._pars2csv()

    def hidden_function(self):
        """ Yaha! You found me!
        """
        print("                  @*****&@         @@.                    ")
        print("           @@@((@ @*******@     @%((((@@((@               ")
        print("         @(((((((@(@*******@@@((@(((((((((@*              ")
        print("        @((((((@(((@*******@(((@%(((((&%(((@              ")
        print("         #@((@((&@((&*******@((@(((((@#(((@               ")
        print("        @@*****@((#@@********@@(((((@(((@/**@@            ")
        print("        @********@((@@*************@@@#@******&%          ")
        print("@@  @ @&(@(***,*******,*******,*******,*******,@      .@@*")
        print(" @   @@((((@*************.*********...*******@(@      ./  ")
        print(" @     %(((#@**********...*,.. ........**,**(@(#*     @   ")
        print("  @      @@,.*.***,@@@.......*,...@@@..*,.....@@    &@    ")
        print("   @(    @....,,...@@......... ,*.@%...*......,%  @@      ")
        print("       @@@.....,***...............,****.......*@,         ")
        print("        .........*............ ........****..*.@          ")
        print("         #*........*...&       @...............@          ")
        print("         @ *.. ... ..,,..@@@@. ... ... ... ...&           ")
        print("          @.,*...........*,...................@           ")
        print("            @.**.............*,..............@            ")
        print("             .@.**...............**.........@             ")
        print("           @@@%   &@@,........ .......**#@    @@          ")
        print("        @,                 (@@@@@@@@(             @       ")
        print("                                                %@*       ")



# -*- coding: utf-8 -*-
# Name: definition_generator.py
# Authors: Stephan Meighen-Berger
# Constructs a file containing the definitions of the different functions
# Also offers the option to convert the parameter file to a csv file

# Imports
import logging
import inspect
from typing import Dict
import collections
import pandas as pd
import pickle
import pkgutil
# Local imports
from .config import config
from .tracks import Track
from .em_cascades import EM_Cascade
from .hadron_cascades import Hadron_Cascade

_log = logging.getLogger(__name__)


class Definitions_Generator(object):
    """Helps with the construction of a definitions file.

    Parameters
    ----------
    track : Track
        The particle for which the tracks should be generated
    em_cascade : Particle
        The particle for which the tracks should be generated
    hadron_cascades : Hadron_Cascade
        The particle for which the tracks should be generated

    Returns
    -------
    None

    Raises
    ------
    """
    def __init__(
            self,
            track: Track,
            em_cascade: EM_Cascade, hadron_cascade: Hadron_Cascade):
        """Helps with the construction of a definitions file.

        Parameters
        ----------
        track : Track
            The particle for which the tracks should be generated
        em_cascade : Particle
            The particle for which the tracks should be generated
        hadron_cascades : Hadron_Cascade
            The particle for which the tracks should be generated

        Returns
        -------
        None

        Raises
        ------
        """
        if not config["general"]["enable logging"]:
            _log.disabled = True
        self._fname = config["advanced"]["generated definitions"]
        self._lines_to_write = []
        # The tracks
        self._lines_to_write.append(
            "# --------------------------------------------------\n",
        )
        self._lines_to_write.append(
            "# Tracks\n",
        )
        self._lines_to_write.append(
            "# --------------------------------------------------\n",
        )
        for val in track.__dict__.values():
            if callable(val):
                self._lines_to_write.append(
                    inspect.getsource(val)
                )
        # The em cascades
        self._lines_to_write.append(
            "# --------------------------------------------------\n",
        )
        self._lines_to_write.append(
            "# EM Cascades\n",
        )
        self._lines_to_write.append(
            "# --------------------------------------------------\n",
        )
        for val in em_cascade.__dict__.values():
            if callable(val):
                self._lines_to_write.append(
                    inspect.getsource(val)
                )
        # The hadron cascades
        self._lines_to_write.append(
            "# --------------------------------------------------\n",
        )
        self._lines_to_write.append(
            "# Hadronic Cascades\n",
        )
        self._lines_to_write.append(
            "# --------------------------------------------------\n",
        )
        for val in hadron_cascade.__dict__.values():
            if callable(val):
                self._lines_to_write.append(
                    inspect.getsource(val)
                )

    def _write(self):
        """ Write the definitions file

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        with open(self._fname, 'w') as f:
            for line in self._lines_to_write:
                f.write(line)

    def _pars2csv(self):
        """ Converts the calculation parameters to a csv file

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        param_file = pkgutil.get_data(
                __name__,
                "data/%s.pkl" % config["scenario"]["parametrization"]
        )
        params = pickle.loads(param_file)
        # Flatten
        params = self._flatten(params)
        pd.DataFrame.from_dict(
            data=params, orient='index'
            ).to_csv('parameters.csv', header=False)

    def _flatten(self, d: Dict, parent_key='', sep='_'):
        """ Helper function to flatten a dictionary of dictionaries

        Parameters
        ----------
        d : Dict
            The dictionary to flatten
        parent_key : str
            Optional: Key in the parent dictionary
        sep : str
            The seperator used

        Returns
        -------
        flattened_dic : dic
            The flattened dictionary
        """
        items = []
        for k, v in d.items():
            new_key = parent_key + sep + str(k) if parent_key else str(k)
            if isinstance(v, collections.abc.MutableMapping):
                items.extend(self._flatten(v, str(new_key), sep=sep).items())
            else:
                items.append((new_key, v))
        return dict(items)



# -*- coding: utf-8 -*-
# Name: tracks.py
# Authors: Stephan Meighen-Berger
# Constructs a track, defined by the emission of photons

import logging
import numpy as np
import pickle
import pkgutil
from .config import config
# Checking if jax should be used
try:
    import jax.numpy as jnp
except ImportError:
    if config["general"]["jax"]:
        raise ImportError("Jax not found!")


_log = logging.getLogger(__name__)


class Track(object):
    """Constructs the track object.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Raises
    ------
    """
    def __init__(self):
        """Constructs the track.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Raises
        ------
        None
        """
        if not config["general"]["enable logging"]:
            _log.disabled = True
        _log.debug('Constructing a track object')
        self._medium = config["scenario"]["medium"]
        self._n = config["mediums"][self._medium]["refractive index"]
        if config["scenario"]["parametrization"] == "aachen":
            _log.info("Loading the aachen parametrization")
            param_file = pkgutil.get_data(
                    __name__,
                    "data/%s.pkl" % config["scenario"]["parametrization"]
            )
            self._params = pickle.loads(param_file)["track"]
        else:
            raise ValueError("Track parametrization " +
                             config["scenario"]["parametrization"] +
                             " not implemented!")
        if config["general"]["jax"]:
            _log.debug("Using JAX")
            self.additional_track_ratio = (
                self._additional_track_ratio_fetcher_jax
            )
            self.cherenkov_angle_distro = (
                self._symmetric_angle_distro_fetcher_jax
            )
        else:
            _log.debug("Using basic numpy")
            self.additional_track_ratio = self._additional_track_ratio_fetcher
            self.cherenkov_angle_distro = self._symmetric_angle_distro_fetcher

    ###########################################################################
    # Basic numpy
    def _additional_track_ratio_fetcher(
            self, E, interaction: str) -> np.array:
        """ Calculates the ratio between the additional track length
        and the original for a single energy.

        Parameters
        ----------
        E : float/np.array
            The energy of the particle in GeV
        interaction : str
            Name of the interaction

        Returns
        -------
        ratio : np.array
            The resulting ratio
        """
        params = self._params["additional track " + self._medium][
            interaction
        ]
        lambd = params["lambda"]
        kappa = params["kappa"]
        ratio = (
            lambd + kappa * np.log(E)
        )
        return ratio

    def _symmetric_angle_distro_fetcher(
            self,
            phi: np.array, n: float,
            E) -> np.array:
        # TODO: Add asymmetry function
        """ Calculates the symmetric angular distribution of the Cherenkov
        emission for a single energy. The error should lie below 10%

        Parameters
        ----------
        phi : np.array
            The angles of interest in degrees
        n : float
            The refractive index
        E : float/np.array
            The energy of interest

        Returns
        -------
        distro : np.array
            The distribution of emitted photons given the angle. The
            result is a 2d array with the first axis for the angles and
            the second for the energies.
        """
        a, b, c = self._energy_dependence_angle_pars(E)
        distro = np.array([
            (a * np.exp(b * np.abs(
                1. / n - np.cos(np.deg2rad(phi_val)))**c
            ))
            for phi_val in phi
        ])

        return distro

    def _energy_dependence_angle_pars(
            self, E):
        """ Parametrizes the energy dependence of the angular distribution
        parameters

        Parameters
        ----------
        E : float / np.array
            The energies of interest

        Returns
        -------
        a : np.array
            The first parameter values for the given energies
        b : np.array
            The second parameter values for the given energies
        c : np.array
            The third parameter values for the given energies
        """
        params = self._params["angular distribution"]
        a_pars = params["a pars"]
        b_pars = params["b pars"]
        c_pars = params["c pars"]
        a = a_pars[0] * (np.log(E)) * a_pars[1]
        b = b_pars[0] * (np.log(E)) * b_pars[1]
        c = c_pars[0] * (np.log(E)) * c_pars[1]
        return a, b, c

    ###########################################################################
    # JAX
    def _additional_track_ratio_fetcher_jax(
            self, E: float, interaction: str) -> float:
        """ Calculates the ratio between the additional track length
        and the original for a single energy. JAX implementation

        Parameters
        ----------
        E : float
            The energy of the particle in GeV
        interaction : str
            Name of the interaction

        Returns
        -------
        ratio : float
            The resulting ratio
        """
        params = self._params["additional track " + self._medium][
            interaction
        ]
        lambd = params["lambda"]
        kappa = params["kappa"]
        ratio = (
            lambd + kappa * jnp.log(E)
        )
        return ratio

    def _symmetric_angle_distro_fetcher_jax(
            self,
            phi: float, n: float,
            E: float) -> float:
        # TODO: Add asymmetry function
        """ Calculates the symmetric angular distribution of the Cherenkov
        emission for a single energy. The error should lie below 10%.
        JAX implementation

        Parameters
        ----------
        phi : float
            The angles of interest in degrees
        n : float
            The refractive index
        E : float
            The energy of interest

        Returns
        -------
        distro : float
            The distribution of emitted photons given the angle. The
            result is a 2d array with the first axis for the angles and
            the second for the energies.
        """
        a, b, c = self._energy_dependence_angle_pars_jax(E)
        distro = (a * jnp.exp(b * jnp.abs(
                1. / n - jnp.cos(jnp.deg2rad(phi)))**c
            ))
        return distro

    def _energy_dependence_angle_pars_jax(
            self, E: float) -> float:
        """ Parametrizes the energy dependence of the angular distribution
        parameters. JAX implementation

        Parameters
        ----------
        E : jnp.array
            The energies of interest

        Returns
        -------
        a : float
            The first parameter value for the given energy
        b : float
            The second parameter value for the given energy
        c : float
            The third parameter value for the given energy
        """
        params = self._params["angular distribution"]
        a_pars = params["a pars"]
        b_pars = params["b pars"]
        c_pars = params["c pars"]
        a = a_pars[0] * (jnp.log(E)) * a_pars[1]
        b = b_pars[0] * (jnp.log(E)) * b_pars[1]
        c = c_pars[0] * (jnp.log(E)) * c_pars[1]
        return a, b, c



# -*- coding: utf-8 -*-
# Name: hadron_cascades.py
# Authors: Stephan Meighen-Berger
# Constructs a hadron cascade, defined by the emitted photons

import logging
import numpy as np
from scipy.special import gamma as gamma_func
from scipy.interpolate import UnivariateSpline
import pickle
import pkgutil
from .config import config
from .particle import Particle
try:
    import jax.numpy as jnp
    from jax.scipy.stats import gamma as jax_gamma
except ImportError:
    if config["general"]["jax"]:
        ImportError("Jax not found!")


_log = logging.getLogger(__name__)


class Hadron_Cascade(object):
    """Constructs the hadron cascade object.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Raises
    ------
    """
    def __init__(self):
        """Constructs the hadron cascade.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Raises
        ------
        ValueError
            Other distribution type for emission angles is not implemented
        ValueError
            File containing the muon production data was not found
        """
        if not config["general"]["enable logging"]:
            _log.disabled = True
        _log.debug('Constructing a hadron cascade object')
        self._medium = config["mediums"][config["scenario"]["medium"]]
        self._n = self._medium["refractive index"]
        self._radlength = self._medium["radiation length"]
        self._Lrad = self._radlength / self._medium["density"]
        if config["scenario"]["parametrization"] == "aachen":
            _log.info("Loading the aachen parametrization")
            param_file = pkgutil.get_data(
                    __name__,
                    "data/%s.pkl" % config["scenario"]["parametrization"]
            )
            self._params = pickle.loads(param_file)["hadron cascade"]
            muon_data = pkgutil.get_data(
                    __name__,
                    "data/%s_muon_production.pkl" % (
                        config["scenario"]["parametrization"]
                    )
            )
            if muon_data is None:
                raise ValueError("Muon production data not found!")
            self.__muon_prod_dict = pickle.loads(muon_data)
            _log.debug("Constructing spline dictionary")
            self.__muon_prod_spl_pars = {}
            for pdg_id in config["simulation"]["hadron particles"]:
                self.__muon_prod_spl_pars[pdg_id] = {
                    "alpha": UnivariateSpline(
                        self.__muon_prod_dict[pdg_id][0],
                        self.__muon_prod_dict[pdg_id][1],
                        k=1, s=0, ext=3
                    ),
                    "beta": UnivariateSpline(
                        self.__muon_prod_dict[pdg_id][0],
                        self.__muon_prod_dict[pdg_id][2],
                        k=1, s=0, ext=3
                    ),
                    "gamma": UnivariateSpline(
                        self.__muon_prod_dict[pdg_id][0],
                        self.__muon_prod_dict[pdg_id][3],
                        k=1, s=0, ext=3
                    )
                }
        else:
            raise ValueError("Hadronic parametrization " +
                             config["scenario"]["parametrization"] +
                             " not implemented!")
        if config["general"]["jax"]:
            _log.info("Running with JAX functions")
            self.cherenkov_angle_distro = self._symmetric_angle_distro_jax
            self.track_lengths = self._track_lengths_fetcher_jax
            self.em_fraction = self._em_fraction_fetcher_jax
            self.long_profile = self._log_profile_func_fetcher_jax
            self.muon_production = self._muon_production_fetcher_jax
        else:
            _log.info("Running with basic functions")
            self.cherenkov_angle_distro = self._symmetric_angle_distro
            self.track_lengths = self._track_lengths_fetcher
            self.em_fraction = self._em_fraction_fetcher
            self.long_profile = self._log_profile_func_fetcher
            self.muon_production = self._muon_production_fetcher

    ###########################################################################
    # Numpy
    def _track_lengths_fetcher(
            self, E, particle: int):
        """ Parametrization for the energy dependence of the tracks

        Parameters
        ----------
        E : float/np.array
            The energy of interest in GeV
        particle : int
            The particle of interest

        Returns
        -------
        track_length : np.array
            The track lengths for different energies
        track_length_dev : np.array
            The track lengths deviations for different energies
        """
        params = self._params["track parameters"][particle]
        alpha = params["alpha"]
        beta = params["beta"]
        alpha_dev = params["alpha dev"]
        beta_dev = params["beta dev"]
        track_length = alpha * E**beta
        track_length_dev = alpha_dev * E**beta_dev
        return track_length, track_length_dev

    def _em_fraction_fetcher(self, E, particle: int):
        """ Parametrization of the EM contribution in a hadronic shower

        Parameters
        ----------
        E : float/np.array
            The energy of interest in GeV
        particle : int
            The particle of interest

        Returns
        -------
        em_fraction : np.array
            The fraction for the given energies
        em_fraction_sd : np.array
            The standard deviation
        """
        params = self._params["em fraction"][particle]
        Es = params["Es"]
        f0 = params["f0"]
        m = params["m"]
        sigma0 = params["sigma0"]
        gamma = params["gamma"]
        em_fraction = 1. - (1. - f0)*(E / Es)**(-m)
        em_fraction_sd = sigma0 * np.log(E)**(-gamma)
        return em_fraction, em_fraction_sd

    def _log_profile_func_fetcher(
            self, E, z: np.array, particle: Particle,
            ) -> np.array:
        """ Parametrization of the longitudinal profile. This still needs work

        Parameters
        ----------
        E : float/np.array
            The energy of interest in GeV
        z : np.array
            The cascade depth in cm
        particle : Particle
            The particle of interest

        Returns
        -------
        res : np.array
            Is equal to l^(-1) * dl/dt. The result will be 2 dimensional, with
            cm defined along the first axis and energies along the second
        """
        t = z / self._Lrad
        b = self._b_energy_fetcher(particle)
        a = self._a_energy_fetcher(E, particle)
        a = np.array([a]).flatten()
        # gamma.pdf seems far slower than the explicit implementation
        res = np.array([
            b * (
                (t * b)**(a_val - 1.) * np.exp(-(t*b)) / gamma_func(a_val)
            ) for a_val in a
        ])
        return res

    def _a_energy_fetcher(self, E: float, particle: int) -> np.array:
        """ Parametrizes the energy dependence of the a parameter for the
        longitudinal profiles

        Parameters
        ----------
        E : float
            The energy of interest in GeV
        particle : int
            The particle of interest

        Returns
        -------
        a : np.array
            The values for the energies of interest
        """
        params = self._params["longitudinal parameters"][
            particle]
        alpha = params["alpha"]
        beta = params["beta"]
        a = alpha + beta * np.log10(E)
        return a

    def _b_energy_fetcher(self, particle: int) -> np.array:
        """ Parametrizes the energy dependence of the b parameter for the
        longitudinal profiles. Currently assumed to be constant

        Parameters
        ----------
        E : float
            The energy of interest in GeV
        particle : int
            The particle of interest

        Returns
        -------
        b : np.array
            The values for the energies of interest
        """
        params = self._params["longitudinal parameters"][
            particle]
        b = params["b"]
        return b

    def _symmetric_angle_distro(
            self, E,
            phi: np.array, n: float,
            particle: Particle) -> np.array:
        # TODO: Add asymmetry function
        """ Calculates the symmetric angular distribution of the Cherenkov
        emission. The error should lie below 10%

        Parameters
        ----------
        E : float/np.array
            The energy of interest in GeV
        phi : np.array
            The angles of interest in degrees
        n : float
            The refractive index
        particle : Particle
            The particle of interest

        Returns
        -------
        distro : np.array
            The distribution of emitted photons given the angle. The
            result is a 2d array with the first axis for the angles and
            the second for the energies.
        """
        a, b, c, d = self._energy_dependence_angle_pars(E, particle)
        distro = np.array([
            (a * np.exp(b * np.abs(
                1. / n - np.cos(np.deg2rad(phi_val)))**c
            ) + d)
            for phi_val in phi
        ])

        return np.nan_to_num(distro)

    def _energy_dependence_angle_pars(
            self, E, particle: int):
        """ Parametrizes the energy dependence of the angular distribution
        parameters

        Parameters
        ----------
        E : float/np.array
            The energy(ies) of interest
        particle : Particle
            The particle of interest

        Returns
        -------
        a : np.array
            The first parameter values for the given energies
        b : np.array
            The second parameter values for the given energies
        c : np.array
            The third parameter values for the given energies
        d : np.array
            The fourth parameter values for the given energies
        """
        params = config[
            "hadron cascade"
        ]["angular distribution"][particle]
        a_pars = params["a pars"]
        b_pars = params["b pars"]
        c_pars = params["c pars"]
        d_pars = params["d pars"]
        a = a_pars[0] * (np.log(E))**a_pars[1]
        b = b_pars[0] * (np.log(E))**b_pars[1]
        c = c_pars[0] * (np.log(E))**c_pars[1]
        d = d_pars[0] * (np.log(E))**d_pars[1]
        return (
            np.array([a]).flatten(), np.array([b]).flatten(),
            np.array([c]).flatten(), np.array([d]).flatten()
        )

    def _muon_production_fetcher(self, Eprim, Emu, particle: Particle):
        """ Parametrizes the production of muons in hadronic cascades

        Parameters
        ----------
        Eprim : float/np.array
            The energy(ies) of the primary particle
        Emu : float/np.array
            The energy(ies) of the muons
        particle : Particle
            The particle of interest

        Returns
        -------
        distro: float/np.array
            The distribution/value of the produced muons
        """
        # Converting the primary energy to an array
        energy_prim = np.array([Eprim]).flatten()
        # Converting to np.array for ease of use
        energy = np.array([Emu]).flatten()
        alpha, beta, gamma = self._muon_production_pars(energy_prim, particle)
        # Removing too small values
        energy[energy <= 1.] = 0.
        # Removing all secondary energies above the primary energy(ies)
        energy_2d = np.array([
            energy
            for _ in range(len(energy_prim))
        ])
        # Removing energies above the primary
        distro = []
        for id_arr, _ in enumerate(energy_2d):
            energy_2d[id_arr][
                energy_2d[id_arr] > energy_prim[id_arr]
            ] = 0.
            # Removing too large values
            energy[
                energy > (alpha[id_arr] / beta[id_arr])**(-1. / gamma[id_arr])
            ] = 0.
            distro.append(energy_prim[id_arr] * (
                -alpha[id_arr] + beta[id_arr] * (
                    energy_2d[id_arr]**(-gamma[id_arr])
                )
            ))
        distro = np.array(distro)
        distro[distro == np.inf] = 0.
        distro[distro < 0.] = 0.
        return distro

    def _muon_production_pars(self, E, particle: Particle):
        """ Constructs the parametrization values for the energies of interest.

        Parameters
        ----------
        E : float/np.array
            The energy(ies) of interest
        particle: Particle
            The particle of interest

        Returns
        -------
        alpha : float/np.array
            The first parameter values for the given energies
        beta : float/np.array
            The second parameter values for the given energies
        gamma : float/np.array
            The third parameter values for the given energies
        """
        alpha = self.__muon_prod_spl_pars[particle._pdg_id]["alpha"](E)
        beta = self.__muon_prod_spl_pars[particle._pdg_id]["beta"](E)
        gamma = self.__muon_prod_spl_pars[particle._pdg_id]["gamma"](E)
        return alpha, beta, gamma

    ###########################################################################
    # JAX
    def _track_lengths_fetcher_jax(
            self, E: float, particle: int):
        """ Parametrization for the energy dependence of the tracks

        Parameters
        ----------
        E : float
            The energy of interest in GeV
        particle : int
            The particle of interest

        Returns
        -------
        track_length : float
            The track lengths for different energies
        track_length_dev : float
            The track lengths deviations for different energies
        """
        params = self._params["track parameters"][particle]
        alpha = params["alpha"]
        beta = params["beta"]
        alpha_dev = params["alpha dev"]
        beta_dev = params["beta dev"]
        track_length = alpha * E**beta
        track_length_dev = alpha_dev * E**beta_dev
        return track_length, track_length_dev

    def _em_fraction_fetcher_jax(self, E: float, particle: int):
        """ Parametrization of the EM contribution in a hadronic shower

        Parameters
        ----------
        E : float
            The energy of interest in GeV
        particle : int
            The particle of interest

        Returns
        -------
        em_fraction : float
            The fraction for the given energies
        em_fraction_sd : float
            The standard deviation
        """
        params = self._params["em fraction"][particle]
        Es = params["Es"]
        f0 = params["f0"]
        m = params["m"]
        sigma0 = params["sigma0"]
        gamma = params["gamma"]
        em_fraction = 1. - (1. - f0)*(E / Es)**(-m)
        em_fraction_sd = sigma0 * jnp.log(E)**(-gamma)
        return em_fraction, em_fraction_sd

    def _log_profile_func_fetcher_jax(
            self, E: float, z: float, particle: int,
            ) -> float:
        """ Parametrization of the longitudinal profile. This still needs work

        Parameters
        ----------
        E : float
            The energy of interest in GeV
        z : float
            The cascade depth in cm
        particle : int
            The particle of interest

        Returns
        -------
        res : int
            Is equal to l^(-1) * dl/dt.
        """
        t = z / self._Lrad
        b = self._b_energy_fetcher_jax(particle)
        a = self._a_energy_fetcher_jax(E, particle)
        res = jax_gamma.pdf(t * b, a) * b
        return res

    def _a_energy_fetcher_jax(self, E: float, particle: int) -> float:
        """ Parametrizes the energy dependence of the a parameter for the
        longitudinal profiles

        Parameters
        ----------
        E : float
            The energy of interest in GeV
        particle : Particle
            The particle of interest

        Returns
        -------
        a : float
            The values for the energies of interest
        """
        params = self._params["longitudinal parameters"][
            particle]
        alpha = params["alpha"]
        beta = params["beta"]
        a = alpha + beta * jnp.log10(E)
        return a

    def _b_energy_fetcher_jax(self, particle: int) -> int:
        """ Parametrizes the energy dependence of the b parameter for the
        longitudinal profiles. Currently assumed to be constant

        Parameters
        ----------
        particle : int
            The particle of interest

        Returns
        -------
        b : int
            The values for the energies of interest
        """
        params = self._params["longitudinal parameters"][
            particle]
        b = params["b"]
        return b

    def _symmetric_angle_distro_jax(
            self, E: float,
            phi: float, n: float,
            particle: int) -> float:
        # TODO: Add asymmetry function
        """ Calculates the symmetric angular distribution of the Cherenkov
        emission. The error should lie below 10%

        Parameters
        ----------
        E : float
            The energy of interest in GeV
        phi : float
            The angles of interest in degrees
        n : float
            The refractive index
        particle : int
            The particle of interest

        Returns
        -------
        distro : float
            The distribution of emitted photons given the angle. The
            result is a 2d array with the first axis for the angles and
            the second for the energies.
        """
        a, b, c, d = self._energy_dependence_angle_pars_jax(E, particle)
        distro = (
            (a * jnp.exp(b * jnp.abs(
                1. / n - jnp.cos(jnp.deg2rad(phi)))**c
            ) + d)
        )

        return jnp.nan_to_num(distro)

    def _energy_dependence_angle_pars_jax(
            self, E: float, particle: int):
        """ Parametrizes the energy dependence of the angular distribution
        parameters

        Parameters
        ----------
        E : float
            The energy of interest
        particle : iny
            The particle of interest

        Returns
        -------
        a : float
            The first parameter value for the given energy
        b : float
            The second parameter value for the given energy
        c : float
            The third parameter value for the given energy
        d : float
            The fourth parameter value for the given energy
        """
        params = self._params["angular distribution"][particle]
        a_pars = params["a pars"]
        b_pars = params["b pars"]
        c_pars = params["c pars"]
        d_pars = params["d pars"]
        a = a_pars[0] * (jnp.log(E))**a_pars[1]
        b = b_pars[0] * (jnp.log(E))**b_pars[1]
        c = c_pars[0] * (jnp.log(E))**c_pars[1]
        d = d_pars[0] * (jnp.log(E))**d_pars[1]
        return (
            a, b,
            c, d
        )

    def _muon_production_fetcher_jax(
            self, Eprim: float, Emu: float, particle: int
            ) -> float:
        """ Parametrizes the production of muons in hadronic cascades

        Parameters
        ----------
        Eprim : float
            The energy of the primary particle
        Emu : float
            The energy of the muons
        particle : int
            The particle of interest

        Returns
        -------
        distro: float
            The distribution/value of the produced muons
        """
        energy_prim = Eprim
        energy = Emu
        alpha, beta, gamma = self._muon_production_pars_jax(
            energy_prim, particle
        )
        # Removing too small values
        if Emu < 1.:
            return 0.
        # Removing all secondary energies above the primary energy(ies)
        if Emu >= energy_prim:
            return 0.
        # Removing too large values
        if Emu > (alpha / beta)**(-1. / gamma):
            return 0.
        distro = (energy_prim * (
            -alpha + beta * (
                energy**(-gamma)
            )
        ))
        # Removing numerical errors
        if distro == np.inf:
            return 0.
        if distro < 0.:
            return 0.
        return distro

    def _muon_production_pars_jax(self, E: float, particle: int):
        """ Constructs the parametrization values for the energies of interest.

        Parameters
        ----------
        E : float
            The energy(ies) of interest
        particle: int
            The particle of interest

        Returns
        -------
        alpha : float
            The first parameter values for the given energy
        beta : float
            The second parameter values for the given energy
        gamma : float
            The third parameter values for the given energy
        """
        alpha = self.__muon_prod_spl_pars[particle]["alpha"](E)
        beta = self.__muon_prod_spl_pars[particle]["beta"](E)
        gamma = self.__muon_prod_spl_pars[particle]["gamma"](E)
        return alpha, beta, gamma



# -*- coding: utf-8 -*-

from .fennel import Fennel
from .config import config

__all__ = (Fennel, config)

# Version of the fennel package
__version__ = "1.2.2"
__author__ = "Stephan Meighen-Berger"



#!/usr/bin/env python

import pathlib
from setuptools import setup

# Parent directory
HERE = pathlib.Path(__file__).parent

# The readme file
README = (HERE / "README.md").read_text()

setup(
    name="fennel_seed",
    version="1.2.2",
    description="Light-yields for tracks, and cascades",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Stephan Meighen-Berger",
    author_email="stephan.meighenberger@gmail.com",
    url='https://github.com/MeighenBergerS/fennel',
    license="MIT",
    install_requires=[
        "PyYAML",
        "numpy",
        "scipy",
        "pandas"
    ],
    extras_require={
        "interactive": ["nbstripout", "matplotlib", "jupyter", "tqdm"],
        "jax": ["jax[cpu]"]
    },
    packages=["fennel"],
    package_data={'fennel': ["data/*.pkl"]},
    include_package_data=True
)
